use std::{
    error::Error, 
    net::{IpAddr, Ipv4Addr, Ipv6Addr, SocketAddr, SocketAddrV4, SocketAddrV6}, 
    str::FromStr, 
    sync::Arc
};

use tokio::sync::RwLock;

use ignore_result::Ignore;
use tokio::{
    io::{AsyncReadExt, AsyncWriteExt}, 
    net::{TcpListener, TcpStream}
};

use log::info;
use tokio_io_timeout::TimeoutStream;
use tokio_rustls::TlsAcceptor;

use crate::tls::create_server_config;

use super::config::{
    Config,
    SiteConfig,
    IpForwarding
};

pub struct FlowgateServer {
    config: Arc<RwLock<Config>>,
}

struct Connection {
    stream: TcpStream, 
    config: SiteConfig,
    keep_alive: bool, 
    host: String,
}

impl FlowgateServer {
    pub fn new(config: Arc<RwLock<Config>>) -> Self {
        FlowgateServer { config }
    }

    pub async fn start(&self) {
        tokio::spawn({ 
            let config = self.config.clone(); 

            async move { 
                Self::run_http(config).await.ignore(); 
            } 
        });

        tokio::spawn({ 
            let config = self.config.clone(); 

            async move { 
                Self::run_https(config).await.ignore(); 
            } 
        });
    }

    pub async fn run_http(
        config: Arc<RwLock<Config>>
    ) -> Result<(), Box<dyn Error>> {
        let listener = TcpListener::bind(&config.read().await.http_host).await?;

        info!("HTTP server runned on {}", &config.read().await.http_host);

        loop {
            let Ok((stream, addr)) = listener.accept().await else { break }; 

            let config = config.clone();

            tokio::spawn(async move {
                let mut stream = TimeoutStream::new(stream);

                stream.set_write_timeout(Some(config.read().await.connection_timeout));
                stream.set_read_timeout(Some(config.read().await.connection_timeout));

                let mut stream = Box::pin(stream);

                Self::accept_stream(
                    config,
                    &mut stream,
                    addr,
                    false
                ).await;
            });
        }

        Ok(())
    }

    pub async fn run_https(
        config: Arc<RwLock<Config>>
    ) -> Result<(), Box<dyn Error>> {
        let listener = TcpListener::bind(&config.read().await.https_host).await?;
        let acceptor = TlsAcceptor::from(Arc::new(create_server_config(config.clone()).await));

        info!("HTTPS server runned on {}", &config.read().await.http_host);

        loop {
            let Ok((stream, addr)) = listener.accept().await else { break }; 

            let config = config.clone();
            let acceptor = acceptor.clone();

            tokio::spawn(async move {
                let mut stream = TimeoutStream::new(stream);

                stream.set_write_timeout(Some(config.read().await.connection_timeout));
                stream.set_read_timeout(Some(config.read().await.connection_timeout));

                let Ok(mut stream) = acceptor.accept(Box::pin(stream)).await else { return };

                Self::accept_stream(
                    config,
                    &mut stream,
                    addr,
                    false
                ).await;
            });
        }

        Ok(())
    }

    pub async fn accept_stream(
        config: Arc<RwLock<Config>>, 
        stream: &mut (impl AsyncReadExt + AsyncWriteExt + Unpin), 
        addr: SocketAddr,
        https: bool
    ) -> Option<()> {
        let mut conn = read_request(config.clone(), stream, addr, https, None).await?;

        if conn.keep_alive && conn.config.enable_keep_alive {
            loop {
                if !conn.config.support_keep_alive {
                    conn.stream.shutdown().await.ignore();
                    conn.stream = conn.config.connect().await?;
                }
                conn = read_request(config.clone(), stream, addr, https, Some(conn)).await?;
            }
        }

        conn.stream.shutdown().await.ignore();
        stream.shutdown().await.ok()?;

        Some(())
    }
}

async fn read_request(
    config: Arc<RwLock<Config>>, 
    stream: &mut (impl AsyncReadExt + AsyncWriteExt + Unpin), 
    addr: SocketAddr,
    https: bool,
    conn: Option<Connection>
) -> Option<Connection> {
    let mut addr = addr;

    match &config.read().await.incoming_ip_forwarding {
        IpForwarding::Simple => {
            let mut header = Vec::new();

            {
                let mut buf = [0; 1];

                while let Ok(1) = stream.read(&mut buf).await {
                    let byte = buf[0];
                    if byte == b'\n' { break }
                    header.push(byte);
                }
            }

            addr = SocketAddr::from_str(&String::from_utf8(header).ok()?).ok()?;
        },
        IpForwarding::Modern => {
            let mut ipver = [0; 1];
            stream.read(&mut ipver).await.ok()?;
            addr = match ipver[0] {
                0x01 => {
                    let mut octets = [0; 4];
                    stream.read(&mut octets).await.ok()?;
                    let mut port = [0; 2];
                    stream.read(&mut port).await.ok()?;
                    let port = u16::from_be_bytes(port);
                    SocketAddr::V4(SocketAddrV4::new(Ipv4Addr::from(octets), port))
                }, 0x02 => {
                    let mut octets = [0; 16];
                    stream.read(&mut octets).await.ok()?;
                    let mut port = [0; 2];
                    stream.read(&mut port).await.ok()?;
                    let port = u16::from_be_bytes(port);
                    SocketAddr::V6(SocketAddrV6::new(Ipv6Addr::from(octets), port, 0, 0))
                }, _ => { return None },
            };
        },
        _ => { }
    }

    let mut head = Vec::new();

    {
        let mut buf = [0; 1];
        let mut counter = 0;

        while let Ok(1) = stream.read(&mut buf).await {
            let byte = buf[0];
            head.push(byte);

            counter = match (counter, byte) {
                (0, b'\r') => 1,
                (1, b'\n') => 2,
                (2, b'\r') => 3,
                (3, b'\n') => break,
                _ => 0,
            };
        }

        head.truncate(head.len() - 4);
    }

    if head.is_empty() { return None; }
    
    let head_str = String::from_utf8(head.clone()).ok()?;
    let head_str = head_str.trim_matches(char::from(0)).to_string();

    let mut head_lines = head_str.split("\r\n");

    let status = head_lines.next()?;
    let status_seq: Vec<&str> = status.split(" ").collect();

    let headers: Vec<(&str, &str)> = head_lines
        .filter(|l| l.contains(": "))
        .map(|l| l.split_once(": ").unwrap())
        .collect();

    let is_chunked = headers.iter()
        .find(|o| o.0.to_lowercase() == "transfer-encoding")
        .map(|o| o.1.split(",").map(|x| x.trim_matches(' ').to_string()).collect::<Vec<String>>())
        .map(|o| o.contains(&"chunked".to_string()))
        .unwrap_or(false);
    
    if let IpForwarding::Header(header) = &config.read().await.incoming_ip_forwarding {
        if let Some(ip) = headers.iter().find(|o| o.0 == header).map(|o| o.1) {
            addr = SocketAddr::from_str(ip).ok()?;
        }
    }

    let mut conn: Connection = if conn.is_none() {
        let mut host = String::new();
        let mut keep_alive = false;

        for (key, value) in &headers {
            match key.to_lowercase().as_str() {
                "host" => host = value.to_string(),
                "connection" => keep_alive = *value == "keep-alive",
                _ => {}
            }
        }

        let site = config.read().await.get_site(&host)?.clone();

        Connection {
            stream: site.connect().await?,
            config: site,
            keep_alive,
            host
        }
    } else {
        conn?
    };
    
    let content_length = headers
        .iter()
        .filter(|(k, _)| k.to_lowercase() == "content-length")
        .next()
        .map(|o| o.1.parse().ok())
        .flatten()
        .unwrap_or(0usize);

    let mut reqbuf: Vec<u8> = Vec::new();

    if let Some(replace_host) = conn.config.replace_host.clone() {
        let mut new_head = Vec::new();
        let mut is_status = true;

        for line in head_str.split("\r\n") {
            if is_status {
                new_head.append(&mut line.as_bytes().to_vec());
                is_status = false;
            } else {
                new_head.append(&mut b"\r\n".to_vec());
                let (key, _) = line.split_once(": ")?;
                if key.to_lowercase() == "host" {
                    new_head.append(&mut key.as_bytes().to_vec());
                    new_head.append(&mut b": ".to_vec());
                    new_head.append(&mut replace_host.as_bytes().to_vec());
                } else {
                    new_head.append(&mut line.as_bytes().to_vec());
                }
            }
        }

        head = new_head;
    }

    match &conn.config.ip_forwarding {
        IpForwarding::Header(header) => {
            reqbuf.append(&mut status.to_string().as_bytes().to_vec());
            reqbuf.append(&mut b"\r\n".to_vec());
            for (key, value) in String::from_utf8(head.clone()).ok()?
                                        .split("\r\n")
                                        .skip(1)
                                        .filter_map(|o| o.split_once(": ")) {
                if *key.to_lowercase() == header.to_lowercase() { continue }
                reqbuf.append(&mut key.to_string().as_bytes().to_vec());
                reqbuf.append(&mut b": ".to_vec());
                reqbuf.append(&mut value.to_string().as_bytes().to_vec());
                reqbuf.append(&mut b"\r\n".to_vec());
            }
            reqbuf.append(&mut header.as_bytes().to_vec());
            reqbuf.append(&mut b": ".to_vec());
            reqbuf.append(&mut addr.to_string().as_bytes().to_vec());
            reqbuf.append(&mut b"\r\n\r\n".to_vec());
        },
        IpForwarding::Simple => {
            reqbuf.append(&mut addr.to_string().as_bytes().to_vec());
            reqbuf.push(b'\n');
            reqbuf.append(&mut head.clone());
            reqbuf.append(&mut b"\r\n\r\n".to_vec());
        },
        IpForwarding::Modern => {
            reqbuf.push(if addr.is_ipv4() { 0x01 } else { 0x02 });
            match addr.ip() {
                IpAddr::V4(ip) => {
                    reqbuf.append(&mut ip.octets().to_vec());
                }, IpAddr::V6(ip) => {
                    reqbuf.append(&mut ip.octets().to_vec());
                }
            }
            reqbuf.append(&mut addr.port().to_be_bytes().to_vec());
            reqbuf.append(&mut head.clone());
            reqbuf.append(&mut b"\r\n\r\n".to_vec());
        },
        IpForwarding::None => {
            reqbuf.append(&mut head.clone());
            reqbuf.append(&mut b"\r\n\r\n".to_vec());
        }
    }

    conn.stream.write_all(&reqbuf).await.ok()?;

    if content_length > 0 {
        let mut read = 0usize;
        let mut buf = vec![0; 4096];
        while let Ok(size) = stream.read(&mut buf).await {
            if size == 0 { break }
            read += size;
            buf.truncate(size);
            conn.stream.write_all(&buf).await.ok()?;
            buf = vec![0; 4096];
            if read >= content_length { break }
        }
    } else if is_chunked {
        loop {
            let mut length = Vec::new();
            {
                let mut buf = [0; 1];
                let mut counter = 0;

                while let Ok(1) = stream.read(&mut buf).await {
                    let byte = buf[0];
                    length.push(byte);

                    counter = match (counter, byte) {
                        (0, b'\r') => 1,
                        (1, b'\n') => break,
                        _ => 0,
                    };
                }
                conn.stream.write_all(&length).await.ok()?;

                length.truncate(length.len() - 2);
            }
            let length = String::from_utf8(length).ok()?;
            let length = usize::from_str_radix(length.as_str(), 16).ok()?;
            let mut data = vec![0u8; length+2];
            stream.read_exact(&mut data).await.ok()?;

            conn.stream.write_all(&data).await.ok()?;
            if length == 0 {
                break;
            }
        }
    }

    if conn.config.support_keep_alive {
        let mut head = Vec::new();

        {
            let mut buf = [0; 1];
            let mut counter = 0;

            while let Ok(1) = conn.stream.read(&mut buf).await {
                let byte = buf[0];
                head.push(byte);

                stream.write_all(&buf).await.ok()?;

                counter = match (counter, byte) {
                    (0, b'\r') => 1,
                    (1, b'\n') => 2,
                    (2, b'\r') => 3,
                    (3, b'\n') => break,
                    _ => 0,
                };
            }

            head.truncate(head.len() - 4);
        }

        if head.is_empty() { return None; }
        
        let head_str = String::from_utf8(head.clone()).ok()?;
        let head_str = head_str.trim_matches(char::from(0));

        let headers = head_str.split("\r\n")
            .skip(1)
            .filter(|l| l.contains(": "))
            .map(|l| l.split_once(": ").unwrap())
            .map(|(k,v)| (k.to_lowercase(),v.to_string()))
            .collect::<Vec<(String,String)>>();

        let content_length = headers.iter()
            .find(|(k, _)| k == "content-length")
            .map(|o| o.1.parse().ok())
            .flatten()
            .unwrap_or(0usize);
        
        let is_chunked = headers.iter()
            .find(|o| o.0.to_lowercase() == "transfer-encoding")
            .map(|o| o.1.split(",").map(|x| x.trim_matches(' ').to_string()).collect::<Vec<String>>())
            .map(|o| o.contains(&"chunked".to_string()))
            .unwrap_or(false);
        
        if content_length > 0 {
            let mut read = 0usize;
            let mut buf = vec![0; 4096];
            while let Ok(size) = conn.stream.read(&mut buf).await {
                if size == 0 { break }
                read += size;
                buf.truncate(size);
                stream.write_all(&buf).await.ok()?;
                buf = vec![0; 4096];
                if read == content_length { break }
            }
        } else if is_chunked {
            loop {
                let mut length = Vec::new();
                {
                    let mut buf = [0; 1];
                    let mut counter = 0;

                    while let Ok(1) = conn.stream.read(&mut buf).await {
                        let byte = buf[0];
                        length.push(byte);

                        counter = match (counter, byte) {
                            (0, b'\r') => 1,
                            (1, b'\n') => break,
                            _ => 0,
                        };
                    }
                    stream.write_all(&length).await.ok()?;

                    length.truncate(length.len() - 2);
                }
                let length = String::from_utf8(length).ok()?;
                let length = usize::from_str_radix(length.as_str(), 16).ok()?;
                let mut data = vec![0u8; length+2];
                conn.stream.read_exact(&mut data).await.ok()?;

                stream.write_all(&data).await.ok()?;
                if length == 0 {
                    break;
                }
            }
        }
    } else {
        let mut buf = vec![0;1024];
        while let Ok(n) = conn.stream.read(&mut buf).await {
            if n == 0 { break }
            buf.truncate(n);
            stream.write_all(&buf).await.ok()?;
            buf = vec![0;1024];
        }
    }

    info!("{addr} > {} {}://{}{}", status_seq[0], if https { "https" } else { "http" }, conn.host, status_seq[1]);

    Some(conn)
}